/* Descriptive model of Arabidopsis thaliana development
 *
 * For a high level description, see "Quantitative modelling of
 * Arabidopsis thaliana development", by Lars Muendermann,
 * Yvette Erasmus, Brendan Lane, Enrico Coen, and Przemyslaw Prusinkiewicz.
 *
 * Copyright 2005 by these authors.
 * University of Calgary, Canada and John Innes Centre, UK
 */

#include <stdio.h>
#include <math.h>

// Useful function definitions
template<typename type>
inline type Max(type x, type y)
{ return (x>y) ? x : y; }

template<typename type>
inline type Min(type x, type y)
{ return (x>y) ? y : x; }

// Include the definition of the MTGParser class
#include "MTGParser.C"
#include <string>

// Include the definitions specific to L+C
#include <lpfgall.h>

using namespace std;

/* Utility functions : sigmoid, allometry, Rad2Deg, Deg2Rad */

// Boltzmann growth function
float sigmoid(float t, float* ap)
{
  return (*ap - *(ap+1)) / (1 + exp ( (t - *(ap+2)) / *(ap+3) ) ) + *(ap+1) ;
}

// Allometry function: used to calculate internode, pedicel, stamen and carpel width
// based on length.
float allometry(float length, float a, float b)
{
  return pow(10.0f, (b*log10(length)+a));
}

// Functions converting between radians and degrees
float Rad2Deg(float r)
{ return r*180.0f/M_PI; }
float Deg2Rad(float r)
{ return r*M_PI/180.0f; }

/* Global constants */

#define TO_DISPLAY 0	// Selects what to display:

#if TO_DISPLAY == 1
#define LEAF		// TO_DISPLAY = 1 displays a single leaf from the main axis
#elif TO_DISPLAY == 2
#define LEAVES		// TO_DISPLAY = 2 displays all of the leaves on the main axis
#elif TO_DISPLAY == 3
#define FLOWER		// TO_DISPLAY = 3 displays a single flower
#else
#define PLANT		// Any other value of TO_DISPLAY displays the entire plant
#endif

#define TT 600		// total growth time [HFS]
const int T0 =  50;	// start time [HFS]
const float DT = 1;	// time increment per simulation step [hours]

const float DS = 0.4f;	// maximum length of (generalized) cylinder segments [mm]
const float SV = 1.0f;	// global length unit [mm]

const float PR = 7.0f;	// carpel length at which petals, sepals, and stamens are dropped

const float RT = 112.02f;	// time at which first flower on the plant begins to form [hours]
const int PL   = 12;	// plastochron for successive flowers [hours]

const int DN = 12;	// maximum number of axes (main + first-order laterals)
const int MN = (19+1);	// maximum number of metamers composing an axis
const int GP =  4;	// number of parameters of the Boltzmann fitting function

const int LN =  4;	// maximum number of recorded developmental stages of a leaf
const int SN =  5;	// maximum number of recorded developmental stages of a sepal
const int PN =  5;      // maximum number of recorded developmental stages of a petal

#ifdef LEAF
#define Y 7	    	// metamer drawn if LEAF option is chosen
#endif
              
#ifdef LEAVES

#define LEAVES_DISPLAY 1	// Selects how the array of leaves is to be displayed:

#if LEAVES_DISPLAY == 1
#define WIDTH			// If LEAVES_DISPLAY = 1, all leaves are scaled to maintain constant width
#else
#define TIME			// Otherwise, the growth of leaves is displayed
#endif

#define X 6		// Constant leaf width if LEAVES are displayed by WIDTH

#endif

/* Position-independent model parameters and functions */

// internode length and width (supporting flowers)
float xl[GP] = {0, 10.6 , 358.2 , 16.8 };
const float XA = 137.5f;
float XL(float t) 
{ 
  return sigmoid (t, &xl[0]); 
}

// pedicel length
float dl[GP] = {0, 13.0, 323, 17.0   };
#define DA(t) (((0-1)/(1+exp((t-dl[2]+300)/dl[3]))+1)*45+20)
float DL(float t) 
{ 
  return sigmoid (t, &dl[0]); 
}

// allometric parameters relating log pedicel width to log pedicel length
// dwAlloA = intercept, dwAlloB = slope
const float dwAlloA = -0.616;
const float dwAlloB = 0.467;

// sepal width
float sw[GP] = {0,  1.3, 206,  78};
float SW(float t)
{
  return sigmoid (t, &sw[0]); 
}

// petal width
float pw[GP] = {0,  1.6, 281,  49 };
float PW(float t)
{
  return sigmoid (t, &pw[0]); 
}

// anther length
float al[GP] = {0,  2.9, 290,  16 };
float AL(float t)
{
  return sigmoid (t, &al[0]); 
}

// allometric parameters relating log stamen filament width to log carpel length
//  awAlloA = intercept, awAlloB = slope
const float awAlloA = -0.7793;
const float awAlloB = 0.3567;

// carpel length
float cl[GP] = {0, 12 , 351 ,  49};
float CL(float t)
{
  return sigmoid (t, &cl[0]); 
}

// allometric parameters relating log carpel width to log carpel length
//  cwAlloA = intercept, cwAlloB = slope
const float cwAlloA = -0.555;
const float cwAlloB = 0.613;

// index of first sepal contour
int sc[ 1] =
   {66};
// widths of measured sepals
float ss[ 1][SN] =
   {0.2801, 0.4057, 0.6111, 0.8639, 1.10};

// index of first petal contour
int pc[ 1] =
   {76};
// widths of measured petals
float ps[ 1][PN] =
   {0.0892, 0.1977, 0.4561, 0.6317, 1.090};

// index of carpel contour
int cc[ 1] =
   {86};

// allometric parameters relating log internode width to log internode length
//  IWAlloA = intercept, IWAlloB = slope
const double IWAlloA = -0.2;
const double IWAlloB = 0.2; 

/* Position-dependent values: will be read in from MTG */

int RM[DN];	// index of first flower-supporting metamer on axis
float RD[DN];	// delay between appearance of first flower on main axis and first flower on current axis

// vegetative internode divergence angle and length
float ia[DN][MN];
float IA(int n,int i)
{
  return ia[n][i];
}

float il[DN][MN][GP];
float IL(int n, int i, float t)
{
  return sigmoid(t,&(il[n][i][0]));
}

// leaf width
float lw[DN][MN][GP];

#if defined(LEAVES) && defined(WIDTH)
#define LW(n,i,t) (X)			// Leaf width is constant in LEAVES mode with WIDTH defined.
#else
float LW(int n, int i, float t)		// Leaf width is computed from the Boltzmann function otherwise.
{
  return sigmoid (t, &lw[n][i][0]); 
}
#endif

// leaf azimuth angle
#define LA(n,i,t) (((0-1)/(1+exp((t-lw[n][i][2])/lw[n][i][3]))+1)*60+30)

// index of first leaf contour of metamer i
int lc[DN][MN];

// widths of measured leaf contours
float ls[DN][MN][LN];

/* Other definitions */

float T;	// Global time (HFS)

// Calculate local time given axis number n and internode number i
float LocalTime(int n, int i)
{
  return T - (i-RM[n])*PL-RT-RD[n];
}

derivation length: int((TT-T0)/DT);

Start:
{
  int i,n;

  /* read MTG */
  ifstream mtgFile("arabidopsis.mtg");
  MTGParser mtg(mtgFile);
  mtgFile.close();

// read metamer parameters (ia,il,lw,lc,ls) from mtg
// Components are identified in the MTG as, for example:
// /A0 is the primary axis
// /A0/M0<M0<M0 is the third metamer of the primary axis (m2)
// /A0/M0<M0<M0/I0+A1 is the axis (subtended by an internode) of that metamer
// /A0/M0<M0<M0/I0+A1/M1<M1 is its second metamer (m2-1).

// Here, we handle the main axis separately from the lateral axes.
// (It was easier this way, because the MTG string prefixes are quite different.)
  string lookStr("/A0/M0");
  for(i = 0 ; i < MN ; i++)
  {
    // mtg.lookupValue takes two parameters: the string representing the topological
    // position of the plant component, and the property name to look up. It returns
    // a C++ string of the value in the MTG file, or an empty string if that property has
    // no value for that component.

    // Note that atof only handles C-style null-terminated strings,
    // so we have to use the C++-string's c_str() method to extract
    // the null-tersinated representation from the C++ representation.

    ia[0][i]    = atof(mtg.lookupValue(lookStr+"/I0","IA").data());

    il[0][i][0] = atof(mtg.lookupValue(lookStr+"/I0","IL0").data());
    il[0][i][1] = atof(mtg.lookupValue(lookStr+"/I0","IL1").data());
    il[0][i][2] = atof(mtg.lookupValue(lookStr+"/I0","IL2").data());
    il[0][i][3] = atof(mtg.lookupValue(lookStr+"/I0","IL3").data());

    lw[0][i][0] = atof(mtg.lookupValue(lookStr+"/I0+L0","LW0").data());
    lw[0][i][1] = atof(mtg.lookupValue(lookStr+"/I0+L0","LW1").data());
    lw[0][i][2] = atof(mtg.lookupValue(lookStr+"/I0+L0","LW2").data());
    lw[0][i][3] = atof(mtg.lookupValue(lookStr+"/I0+L0","LW3").data());

    lc[0][i] = atoi(mtg.lookupValue(lookStr+"/I0+L0","LC").data());

    ls[0][i][0] = atof(mtg.lookupValue(lookStr+"/I0+L0","LS0").data());
    ls[0][i][1] = atof(mtg.lookupValue(lookStr+"/I0+L0","LS1").data());
    ls[0][i][2] = atof(mtg.lookupValue(lookStr+"/I0+L0","LS2").data());
    ls[0][i][3] = atof(mtg.lookupValue(lookStr+"/I0+L0","LS3").data());

    // Here we add another metamer to the string representing the axis,
    // thus iterating over metamers along the main axis.
    lookStr += "<M0";
  }

// Here we read in the data for the lateral axes.
// There are two loops; the outer loop (over n) progresses along the main axis,
// and the inner loop (over i) progresses along the lateral axes.
  lookStr.assign("/A0/M0");
  for(n = 1 ; n < DN ; n++)
  {
    string subStr(lookStr + "/I0+A1/M1");
    for(i = 0 ; i < MN ; i++)
    {
      ia[n][i]    = atof(mtg.lookupValue(subStr+"/I1","IA").data());

      il[n][i][0] = atof(mtg.lookupValue(subStr+"/I1","IL0").data());
      il[n][i][1] = atof(mtg.lookupValue(subStr+"/I1","IL1").data());
      il[n][i][2] = atof(mtg.lookupValue(subStr+"/I1","IL2").data());
      il[n][i][3] = atof(mtg.lookupValue(subStr+"/I1","IL3").data());

      lw[n][i][0] = atof(mtg.lookupValue(subStr+"/I1+L1","LW0").data());
      lw[n][i][1] = atof(mtg.lookupValue(subStr+"/I1+L1","LW1").data());
      lw[n][i][2] = atof(mtg.lookupValue(subStr+"/I1+L1","LW2").data());
      lw[n][i][3] = atof(mtg.lookupValue(subStr+"/I1+L1","LW3").data());

      lc[n][i] = atoi(mtg.lookupValue(subStr+"/I1+L1","LC").data());

      ls[n][i][0] = atof(mtg.lookupValue(subStr+"/I1+L1","LS0").data());
      ls[n][i][1] = atof(mtg.lookupValue(subStr+"/I1+L1","LS1").data());
      ls[n][i][2] = atof(mtg.lookupValue(subStr+"/I1+L1","LS2").data());
      ls[n][i][3] = atof(mtg.lookupValue(subStr+"/I1+L1","LS3").data());

      subStr += "<M1";
    }

    lookStr += "<M0";
  }

// read axial parameters (RM and RD) from mtg
// These are attributes of whole axes rather than individual metamers,
// and are easier to read in separately.
  RM[0] = atoi(mtg.lookupValue("/A0","RM").data());
  RD[0] = atoi(mtg.lookupValue("/A0","RD").data());
  lookStr.assign("/A0/M0");
  for(i = 1 ; i < DN ; i++)
  {
    RD[i] = atof(mtg.lookupValue(lookStr+"/I0+A1","RD").data());
    RM[i] = atoi(mtg.lookupValue(lookStr+"/I0+A1","RM").data());
    lookStr += "<M0";
  }


  // set global time 'T' to 'T0'
  T=T0;
}

StartEach:
// increment global time 'T' by 'DT' after every step
{ T += DT; }

/* Module declarations */

module LeafArray();		// module for visualizing the array of leaves
module Time();

module Apex(int,int,int);
module Internode(int, int, int);
module Leaf(int, int);
module Flower(int, int);
module Pedicel(float);
module Sepal(float);
module Petal(float);
module Stamen(float);
module Carpel(float);

/* The L-system rules */

interpretation:
maximum depth: 1000;
  
// list of possible axioms

#ifdef LEAVES
axiom: SetColor(59) LeafArray();
#endif  

#ifdef LEAF
axiom: SetColor(59) Leaf(0,Y);
#endif  

#ifdef FLOWER
axiom: SetColor(33) Flower(0,RM[0]);
#endif 

#ifdef PLANT
axiom: RollR(166) SetColor(59) Elasticity(0.12) Apex(0,0,0);
#endif 


/* Draws global time on screen */
Time() :
{
  static char bf[32];
  sprintf(bf, "HFS %.2f", T);
  produce SB SetColor(60) Left(180) f(10*SV) Label(bf) EB;
}

#ifdef LEAVES
// Visualize leaves as a 2D array
LeafArray() :
{
  static char tmp[3][3][5];
  for (int row = 0; row < 3; row++)
    for (int column = 0; column < 3; column++)
    {
      sprintf(tmp[row][column],"m%d\0",row*3 + column + 2);
      nproduce SB MoveTo(50*column, -60*row, 0)
                  SetColor(59) Leaf(0, row*3+column+2) EB ;
      nproduce SB MoveTo(1+50*column, 1-60*row, 2)
                  SetColor(2) Label((const char*)tmp[row][column]) EB;
    }
  produce ;
}
#endif   /* LEAVES */


#ifdef PLANT
// recursively create the branching structure according to plant age
Apex(o, n, i):
{
  if ( o >= 2 /*|| n >= DN*/ )
    produce ;

  if (i < RM[n])  // metamer that supports leaves
    produce RollL(IA(n,i))  Left(4) Internode(0, n, i )
            SB Elasticity(0) Leaf(n,i) EB
            SB Left(33) Apex(o+1,i+1,0) EB
            Apex(o,n,i+1); 

  float t = LocalTime( n, i); 
  if (
      RM[n] > 0 &&             // metamer that supports flowers
      (IL(n,RM[n],t) > 0.05    // internode length > threshold
       || DL(t) > 0.05)        //  pedicel length > threshold
     )
    produce RollL(XA) Left(4) Internode(1 ,n, i) SB Flower(n,i) EB Apex(o,n,i+1);      
}
#endif /* PLANT */

/* Visualize an internode */
Internode(c,n,i):
{
  const float MIN_INTERNODE_SEGMENTS = 3;

  float width, length;
 
  if (c == 0 && il[n][i][3] != 0 ) 
  {
    length = IL(n,i,T);
  }
  else if (c == 1)
  {
    float t = LocalTime( n, i);
    length = XL(t);
  }
  else 
    produce ;
  
  // calculates width based on allometric parameters
  width = allometry(length, IWAlloA, IWAlloB);

  float ns = Max(MIN_INTERNODE_SEGMENTS, length/DS); 
  nproduce Elasticity(0.12/ns) SetWidth(width);
  float l = length*SV/(ns-1);
  for (float j=0; j<ns; j+=1)
    nproduce F(l);
  produce;
}

// Pair of contour indices for interpolation
struct IndexPair 
{
  int smaller;
  int larger;
};

/* Find two indices, ip.smaller and ip.larger, of the organs (leaves, petals, sepals...) 
   to be interpolated.  If no interpolation is to be performed, return ip.smaller equal
   to ip.larger. */
IndexPair FindIndexPair(float width, float arr[], int max_number)
{
  /* find the widest measured organ (leaf, petal, sepal...) less wide than "width".  
     Ignore zeroes padding the array of measured widths "arr". */
  int l = 0;
  for (int j=0; j<max_number; ++j)
    if ((arr[j] <= width) && (arr[j] > 0))
      l=j;

  IndexPair ip;
  /* Interpolate if arr[l] <= width < arr[l+1], don't interpolate
     otherwise. */
  if ( (l < max_number-1) && (width < arr[l+1]) )
  { ip.smaller = l; ip.larger = l+1; }
  else
  { ip.smaller = ip.larger = l; }

  return ip;
}

/* Find coefficient r to determine where "current" point lies between "smaller"
   and "larger" points for the subsequent interpolation purposes */

float FindWeight(float current, float smaller, float larger) 
{
  return(current-smaller)/(larger-smaller);
}


/* Interpolate points lying on two curves */

V2f InterpolateCurves (int first_curve, IndexPair ip, int step, float r, float s)
/* first_curve: index of the first curve in the sequence from which the curves
   	to be interpolated are drawn.
   ip: the IndexPair indicating the indices of the curves to be interpolated.
   step: the step between two curves of the same type: step = 1 if all curves
   	in the sequence pointed to by first_curve are of the same type (e.g., leaf 
	contours), 2 if these curves are grouped in pairs (contour - bending),
	which is the case for sepals and petals.
   r:   weight with which the two curves will be combined.
   s:   arc-length coordinate of the query point on each curve, normalized to 0-1. */

{
  if (ip.smaller != ip.larger)
    return (1-r)*curveXY(first_curve+step*ip.smaller, s) +
              r *curveXY(first_curve+step*ip.larger,  s);
  else
    return curveXY(first_curve+step*ip.smaller, s);
}


/* visualize leaf on axis n, index i, at time t */
Leaf(n, i) :
{
  const float MIN_LEAF_SEGMENTS = 3;

  if (i>=RM[n] || (lw[n][i][3] == 0))	// if index i outside the vegatative zone...
    produce; 				// ...don't visualize
   
  /* find the size of leaf on axis n, internode i, at current time */
  float current_width = LW(n,i,T);
   
  /* find the two leaf shapes (l1,l2) that are interpolated to obtain
     the current leaf shape */ 
  IndexPair ip = FindIndexPair(current_width, &ls[n][i][0], LN);

  /* find the weight r for interpolating between two measured contours */
  float r = FindWeight(current_width, ls[n][i][ip.smaller], ls[n][i][ip.larger]);

  /* calculate step size along leaf silhouette as a function of leaf size (length) */
  float y_length = InterpolateCurves(lc[n][i], ip, 1, r, 1).y;
  const float DrawStep = 0.1f;
  float ns = Max(MIN_LEAF_SEGMENTS, y_length / DrawStep);
  float ds = 1/(ns-1);

  /* initialize generalized cylinder */
  float x_init = InterpolateCurves(lc[n][i],ip,1,r,0).x;
  nproduce Down(-LA(n,i,T)) CurrentContour(leaf) SetWidth(-2*current_width*x_init*SV) StartGC;

  /* traverse leaf contour with step ds, and add consecutive segments 
     to the generalized cylinder */
  float s=0;		// percentage of silhouette travelled
  float old_y=0; 	// old_y position, used to calculate y increments
  while (s<1) 
  {
    s = Min(1.0f, s+ds);
    V2f curve_point = InterpolateCurves(lc[n][i], ip, 1, r, s);
    float dy = curve_point.y-old_y;
    nproduce SetWidth(-2*current_width*curve_point.x*SV) F(dy*current_width*SV) 
             Down(-2*0.25*ds/0.01); 
    old_y = curve_point.y;
  }
   
  /* terminate the generalized cylinder */
  produce EndGC;
}

/* Flower: a structure consisting of pedicel, sepals, petals, carpel */
Flower(n,i) :
{
  if (i<RM[n])
    produce ;

  float t = LocalTime( n, i);

  nproduce Pedicel(t);
  nproduce SB Carpel(t) EB;

  // The rest of the organs are dropped if the carpel length
  // exceeds the flower-drop length PR:
  if(CL(t) < PR)
  {
    // Four sepals, separated by 90 degrees
    nproduce SB SetColor(59);
    for(int c = 0 ; c < 4 ; c++)
      nproduce SB Sepal(t) EB RollL(90);
    nproduce EB;

    // Four petals, separated by 90 degrees
    nproduce SB SetColor(107) RollL(45);
    for(int c = 0 ; c < 4 ; c++)
      nproduce SB Petal(t) EB RollL(90);
    nproduce EB;

    // Six stamen filaments
    nproduce SB SetColor(111) RollL(45);
    for(int c = 0 ; c < 4 ; c++)
      nproduce SB Left(25) Stamen(t) EB RollL(90);
    nproduce EB;

    nproduce SB SetColor(111) RollL(90);
    for(int c = 0 ; c < 2 ; c++)
      nproduce SB Left(25) Stamen(t) EB RollL(180);
    nproduce EB;
  }

  produce;
}

/* visualize pedicel as numerous segments */
Pedicel(t) :
{
  const float MIN_PEDICEL_SEGMENTS = 5;

  float da = DA(t) , dl = DL(t) , dw = allometry(dl, dwAlloA, dwAlloB); 
  float ns = Max(MIN_PEDICEL_SEGMENTS, dl/DS);

  nproduce SetColor(59) Left(da) SetWidth(dw*SV) Elasticity(0.62/ns); 
  for (float j=0; j<ns; j+=1)
    nproduce F(dl*SV/(ns-1));
  produce f(-dw*SV) Elasticity(0);
}

/* visualize sepal */
Sepal(t) :
{
  const float MIN_SEPAL_SEGMENTS = 6;

  float sw = SW(t);

  /* find the two sepal shapes (ip.smaller, ip.larger) that are interpolated 
     to obtain current sepal shape */ 
  IndexPair ip = FindIndexPair(sw, &ss[0][0], SN);

  /* find the weight r for interpolating between two measured contours */
  float r = FindWeight(sw, ss[0][ip.smaller], ss[0][ip.larger]);
  
  /* calculate step size along sepal silhouette as a function of sepal length */
  float y_length = InterpolateCurves(sc[0], ip, 2, r, 1).y;
  float ns = Max(MIN_SEPAL_SEGMENTS, y_length * sw / DS);
  float ds = 1/(ns-1);

  /* initialize generalized cylinder */
  float x_init = InterpolateCurves(pc[0],ip,2,r,0).x;
  nproduce CurrentContour(sepal) RollR(90) SetWidth(-2*sw*x_init*SV) StartGC RollR(-90); 

  /* traverse sepal contour with step ds, and add consecutive segments 
     to the generalized cylinder */

  float s = 0;		// percentage of silhouette travelled
  float old_y = 0;	// old_y position, used to calculate y increments
  V2f OldPoint(0.0, 0.0); // old point
  V2f OldHead(0.0, 1.0);  // old heading (orientation) for bending calculcations
 
  while (s<1) 
  {
    s = Min(s+ds,1.0f);
    V2f curve_point = InterpolateCurves(sc[0], ip, 2, r, s);

    
    /* calculate bending angle alpha */
    // q = relative position along the midrib, for curvature computation
    float q = Min(curve_point.y/y_length, 1.0f); 
    V2f xypoint = InterpolateCurves(sc[0]+1, ip, 2, r, q);

    V2f tangent = xypoint - OldPoint;       	
    tangent.Normalize();
     
    float theta = OldHead*tangent;  // dot product
    if (theta<-1)
      theta = -1;
    else if (theta>1)
      theta = 1;
    float alpha = Rad2Deg(acos(theta));
    if (tangent.x > OldHead.x)
      alpha = -alpha;
    /* end of the calculatation of the bending angle alpha */


    // step along the y axis (midrib) of the sepal
    float dy = curve_point.y-old_y;
    nproduce Right(alpha) SetWidth(-2*sw*curve_point.x*SV)
             RollR(90) F(dy*sw*SV) RollR(-90) ;
    old_y = curve_point.y;
    OldPoint = xypoint;
    OldHead = tangent;
  }
    
  produce EndGC;
}

/* visualize petal */
Petal(t) :
{
  const float MIN_PETAL_SEGMENTS = 6;

  float pw = PW(t);

  /* find the two petal shapes (ip.smaller, ip.larger) that are interpolated 
     to obtain current sepal shape */ 
  IndexPair ip = FindIndexPair(pw, &ps[0][0], PN);

  /* find the weight r for interpolating between two measured contours */
  float r = FindWeight(pw, ps[0][ip.smaller], ps[0][ip.larger]);
  
  /* calculate step size along petal silhouette as a function of petal size (length) */
  float y_length = InterpolateCurves(pc[0], ip, 2, r, 1).y;
  float ns =Max(MIN_PETAL_SEGMENTS, y_length*pw/DS);
  float ds = 1/(ns-1);
   
  /* initialize generalized cylinder */
  float x_init = InterpolateCurves(pc[0],ip,2,r,0).x;
  nproduce CurrentContour(petal) SetColor(112) RollR(90) SetWidth(-2*pw*x_init*SV) StartGC RollR(-90); 
   
  /* traverse petal contour with step ds, and add consecutive segments 
     to the generalized cylinder */
  float s = 0;		// percentage of silhouette travelled
  float old_y = 0;	// old_y position, used to calculate y increments
  V2f OldPoint(0.0, 0.0); // old point
  V2f OldHead(0.0, 1.0);  // and and orientations for bending calculaitons
 
  while (s<1) 
  {
    s = Min(s+ds,1.0f);
    V2f curve_point = InterpolateCurves(pc[0], ip, 2, r, s);


    /* calculate bending angle alpha */
    // relative position along the midrib, for curvature computation
    float q = Min(curve_point.y/y_length, 1.0f);
    V2f xypoint = InterpolateCurves(pc[0]+1, ip, 2, r, q);

    V2f tangent = xypoint - OldPoint;
    tangent.Normalize();

    float theta = OldHead * tangent; // dot product
    if (theta < -1)
      theta = -1;
    else if (theta>1)
      theta = 1;
    float alpha=Rad2Deg(acos(theta));
    if (tangent.x>OldHead.x)
      alpha = -alpha;
    /* end of the calculatation of the bending angle alpha */


    // step along the y axis (midrib) of the petal
    float dy = curve_point.y-old_y;
    nproduce Right(alpha) RollR(90) SetWidth(-2*pw*curve_point.x*SV)
            F(dy*pw*SV) PointGC RollR(-90) ;
    old_y = curve_point.y;
    OldPoint = xypoint;
    OldHead = tangent;
  }
  produce EndGC;
}


/* visualize stamen filament */
Stamen(t) :
{
  const float MIN_STAMEN_SEGMENTS = 5;

  float al = AL(t) , aw = allometry(al, awAlloA, awAlloB); 

  float ns = Max(MIN_STAMEN_SEGMENTS, al/DS);

  nproduce SetColor(11)  SetWidth(aw*SV) Elasticity(0.62/ns); 
  for (float j=0; j<ns; j+=1)
    nproduce F(al*SV/(ns-1));
  produce;  //f(-aw*0.2*SV) Sphere(0.6*aw) Elasticity(0);
}


/* visualize carpel */
Carpel(t) :
{
  const float MIN_CARPEL_SEGMENTS = 3;

  float cl = CL(t) , cw = allometry(cl, cwAlloA, cwAlloB); 

  /* calculate the number of segments */
  float ns = Max(MIN_CARPEL_SEGMENTS, cl/DS);
  float ds = 1/(ns-1);
  
  /* initialize the generalized cylinder */
  nproduce SetColor(59) CurrentContour(0) RollR(90) StartGC RollR(-90);

  float s = 0.0;
  float ol = 0.0;

  /* create consecutive segments */
  while (s<1)
  {
    s = Min(1.0f, s+ds); 
    float w = curveX(cc[0], s);
    float l = curveY(cc[0], s);
    
    float dl = l-ol;
    nproduce SetWidth(2*cw*w*SV) RollR(90)F(dl*cl*SV)
             PointGC RollR(-90);
    ol = l;
  }
  produce ;
}
